'''
scisoftpy.pyplot is the implementation of the plotting functionality that delegates
method calls to GDA/SDA via XML-RPC calls.

For SDAPlotter functions that take IDatasets, this wrapper converts numpy.ndarrays to IDatasets
via a temporary file written in NumPy file format.
'''
import numpy as _np #@UnresolvedImport
from xmlrpclib import ServerProxy
from tempfile import mkstemp
import os 
import sys

import scisoftpy.roi as _roi

# @todo this may need configuring
_server = ServerProxy("http://localhost:8610")

def _serialize_npy(array):
    '''Serialize an ndarray to file and return the name of that file'''
    (osfd, filename) = mkstemp(suffix='.npy', prefix='scisofttmp-')
    os.close(osfd)
    try:
        _np.save(filename, array)
        return filename
    except Exception, e:
        # If we failed to write the file, remove it
        # mkstemp returned a new file that did not exist, so 
        # we can't be removing someone else's file
        os.remove(filename)
        raise e
    
def _make_xmlrpc_call(method, *args):
    
    files_to_delete=[]
    try:
        outargs=[]
        for arg in args:
            # @todo on small arrays wrapping the serialized array
            # in Binary may be faster (arrives as byte[] in Java)
            if isinstance(arg, _np.ndarray):
                filename = _serialize_npy(arg)
                files_to_delete.append(filename)
                outargs.append(filename)
            elif isinstance(arg, (list, tuple)):
                filenamelist = []
                for elem in arg:
                    filename = _serialize_npy(elem)
                    filenamelist.append(filename)
                    files_to_delete.append(filename)
                outargs.append(filenamelist)
            else:
                outargs.append(arg)
        # make the xmlrpc call
        return getattr(_server.Plotter, method)(*outargs)
    finally:
        for file_to_delete in files_to_delete:
            try:
                os.remove(file_to_delete)
            except Exception:
                # The XML-RPC server is supposed to return once it is done with the temporary file. However, 
                # due to a bug in the JVM on Windows it can still hold the file after it is complete
                # see http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4715154
                print >> sys.stderr, "Failed to remove temporary file '%s' after completing XML-RPC call" % file_to_delete
                
            
def _unflatten(obj):
    if isinstance(obj, dict):
        if "__type__" in obj:
            if obj["__type__"] == "RectangularROI":
                obj = _roi.rect(obj)
            elif obj["__type__"] == "GuiBean":
                newobj = dict()
                for k, v in obj.items():
                    if (k != "__type__"):
                        newobj[k] = _unflatten(v)
                obj = newobj
    if isinstance(obj, list):
        newobj = []
        for elem in obj:
            newobj.append(_unflatten(elem))
        obj = newobj
    return obj
            
def plot_line(*args):
    '''Python version of SDAPlotter methods:
            plot(String plotName, IDataset yAxis)
            plot(String plotName, IDataset xAxis, IDataset yAxis)
            plot(String plotName, IDataset xAxis, IDataset[] yAxes)
    '''
    _make_xmlrpc_call("plot", *args)
    pass
    
plot_updateline = None # TODO
plot_stack = None # TODO

def plot_image(*args):
    '''Python version of SDAPlotter methods:
            imagePlot(String plotName, IDataset image)
            imagePlot(String plotName, IDataset xAxis, IDataset yAxis, IDataset image)
    '''
    _make_xmlrpc_call("imagePlot", *args)
    
plot_images = None # TODO
plot_surface = None # TODO
plot_points2d = None # TODO
plot_updatepoints2d = None # TODO
plot_points3d = None # TODO
plot_updatepoints3d = None # TODO

plot_scanforimages = None # TODO
plot_viewnexustree = None # TODO
plot_volume = None # TODO

def plot_getbean(*args):
    bean = _make_xmlrpc_call("getGuiBean", *args)
    return _unflatten(bean)

plot_setbean = None # TODO

class parameters:
    plotmode = "PlotMode"
    title = "Title"
    roi = "ROI"
    roilist = "ROIList"
    plotid = "PlotID"
    plotop = "PlotOp"
    fileop = "FileOp"
    filename = "Filename"
    fileselect = "FileList"
    dispview = "DisplayOnView"
    
class imageorder:
    IMAGEORDERNONE = 0
    IMAGEORDERALPHANUMERICAL = 1
    IMAGEORDERCHRONOLOGICAL = 2

# TODO For now guibean is just a dict, but it may become a more typed object
guibean = dict
exception = Exception # TODO

_imageorder = imageorder()
__orders = { "none": _imageorder.IMAGEORDERNONE, "alpha": _imageorder.IMAGEORDERALPHANUMERICAL, "chrono": _imageorder.IMAGEORDERCHRONOLOGICAL}

def order(order):
    try:
        return __orders[order]
    except KeyError:
        raise ValueError, "Given order not one of none, alpha, chrono"

    


